#!/usr/bin/env python

import argparse, psycopg2, time, signal, sys

partman_version = "2.0.0"

parser = argparse.ArgumentParser(description="This script calls either partition_data_time() or partition_data_id() depending on the value given for --type. A commit is done at the end of each --interval and/or fully created partition. Returns the total number of rows moved to partitions. Automatically stops when parent is empty. See docs for examples.", epilog="NOTE: To help avoid heavy load and contention during partitioning, autovacuum is turned off for the entire partition set when this script is run. When partitioning is complete, autovacuum is set back to its default value and the parent table is vacuumed.")
parser.add_argument('-p','--parent', help="Parent table of an already created partition set. (Required)")
parser.add_argument('-t','--type', choices=["time","id",], help="""Type of partitioning. Valid values are "time" and "id". (Required)""")
parser.add_argument('-c','--connection', default="host=", help="""Connection string for use by psycopg. Defaults to "host=" (local socket). Note that two connections are required if allowing autovacuum to be turned off.""")
parser.add_argument('-i','--interval', help="Value that is passed on to the partitioning function as p_batch_interval argument. Use this to set an interval smaller than the partition interval to commit data in smaller batches. Defaults to the partition interval if not given.")
parser.add_argument('-b','--batch', default=0, type=int, help="""How many times to loop through the value given for --interval. If --interval not set, will use default partition interval and make at most -b partition(s). Script commits at the end of each individual batch. (NOT passed as p_batch_count to partitioning function). If not set, all data in the parent table will be partitioned in a single run of the script.""")
parser.add_argument('-w','--wait', default=0, type=float, help="Cause the script to pause for a given number of seconds between commits (batches) to reduce write load")
parser.add_argument('-o', '--order', choices=["ASC", "DESC"], default="ASC", help="Allows you to specify the order that data is migrated from the parent to the children, either ascending (ASC) or descending (DESC). Default is ASC.")
parser.add_argument('-l','--lockwait', default=0, type=float, help="Have a lock timeout of this many seconds on the data move. If a lock is not obtained, that batch will be tried again.")
parser.add_argument('--lockwait_tries', default=10, type=int, help="Number of times to allow a lockwait to time out before giving up on the partitioning.  Defaults to 10")
parser.add_argument('--autovacuum_on', action="store_true", help="Turning autovacuum off requires a brief lock to ALTER the table property. Set this option to leave autovacuum on and avoid the lock attempt.")
parser.add_argument('-q','--quiet', action="store_true", help="Switch setting to stop all output during and after partitioning for use in cron jobs")
parser.add_argument('--version', action="store_true", help="Print out the minimum version of pg_partman this script is meant to work with. The version of pg_partman installed may be greater than this.")
parser.add_argument('--debug', action="store_true", help="Show additional debugging output")
args = parser.parse_args()

is_autovac_off = False

def close_conn(conn):
    conn.close()


def create_conn():
    conn = psycopg2.connect(args.connection)
    conn.autocommit = True
    return conn


def get_partman_schema(conn):
    cur = conn.cursor()
    sql = "SELECT nspname FROM pg_catalog.pg_namespace n, pg_catalog.pg_extension e WHERE e.extname = 'pg_partman' AND e.extnamespace = n.oid"
    cur.execute(sql)
    partman_schema = "\"" + cur.fetchone()[0] + "\""
    cur.close()
    return partman_schema


def get_quoted_parent_table(conn):
    cur = conn.cursor()
    sql = "SELECT schemaname, tablename FROM pg_catalog.pg_tables WHERE schemaname||'.'||tablename = %s"
    cur.execute(sql, [args.parent])
    result = cur.fetchone()
    if result == None:
        print("Given parent table ("+args.parent+") does not exist")
        sys.exit(2)
    quoted_parent_table = "\"" + result[0] + "\".\"" + result[1] + "\""
    cur.close()
    return quoted_parent_table

def partition_data(conn, partman_schema):
    batch_count = 0
    total = 0
    lockwait_count = 0

    cur = conn.cursor()

    sql = "SELECT " + partman_schema + ".partition_data_" + args.type + "(%s"
    if args.interval != "":
        sql += ", p_batch_interval := %s"
    sql += ", p_lock_wait := %s"
    sql += ", p_order := %s"
    sql += ", p_analyze := FALSE)"

    while True:
        if args.interval != "":
            li = [args.parent, args.interval, args.lockwait, args.order]
        else:
            li = [args.parent, args.lockwait, args.order]
        if args.debug:
            print(cur.mogrify(sql, li))
        cur.execute(sql, li)
        result = cur.fetchone()
        if not args.quiet:
            if result[0] > 0:
                print("Rows moved: " + str(result[0]))
            elif result[0] == -1:
                print("Unable to obtain lock, trying again")
                print(conn.notices[-1])
        # if lock wait timeout, do not increment the counter
        if result[0] != -1:
            batch_count += 1
            total += result[0]
            lockwait_count = 0
        else:
            lockwait_count += 1
            if lockwait_count > args.lockwait_tries:
                print("Quitting due to inability to get lock on next rows to be moved")
                break
        # If no rows left or given batch argument limit is reached
        if (result[0] == 0) or (args.batch > 0 and batch_count >= int(args.batch)):
            break
        time.sleep(args.wait)

    return total


def print_version():
    print(partman_version)
    sys.exit()


def reset_autovacuum(partman_schema, quoted_parent_table):
    global is_autovac_off
    vacuum_conn = create_conn()
    cur = vacuum_conn.cursor()
    sql = "ALTER TABLE " + quoted_parent_table + " RESET (autovacuum_enabled, toast.autovacuum_enabled)"
    if not args.quiet:
        print("Attempting to reset autovacuum for old parent table and all child tables...")
    if args.debug:
        print(cur.mogrify(sql))
    cur.execute(sql)
    sql = "SELECT partition_schemaname, partition_tablename FROM " + partman_schema + ".show_partitions(%s)"
    if args.debug:
        print(cur.mogrify(sql, [args.parent]))
    cur.execute(sql, [args.parent])
    result = cur.fetchall()
    for r in result:
        sql = "ALTER TABLE \"" + r[0] + "\".\"" + r[1] + "\" RESET (autovacuum_enabled, toast.autovacuum_enabled)"
        if args.debug:
            print(cur.mogrify(sql))
        cur.execute(sql)
    print("\t... Success!")
    is_autovac_off = False
    cur.close()
    close_conn(vacuum_conn)


def signal_handler(signum, frame):
    if is_autovac_off == True:
        reset_autovacuum(partman_schema, quoted_parent_table)
    sys.exit(2)


def turn_off_autovacuum(partman_schema, quoted_parent_table):
    global is_autovac_off
    vacuum_conn = create_conn()
    cur = vacuum_conn.cursor()
    sql = "ALTER TABLE " + quoted_parent_table + " SET (autovacuum_enabled = false, toast.autovacuum_enabled = false)"
    if not args.quiet:
        print("Attempting to turn off autovacuum for partition set...")
    if args.debug:
        print(cur.mogrify(sql))
    cur.execute(sql)
    sql = "SELECT partition_schemaname, partition_tablename FROM " + partman_schema + ".show_partitions(%s)"
    if args.debug:
        print(cur.mogrify(sql, [args.parent]))
    cur.execute(sql, [args.parent])
    result = cur.fetchall()
    for r in result:
        sql = "ALTER TABLE \"" + r[0] + "\".\"" + r[1] + "\" SET (autovacuum_enabled = false, toast.autovacuum_enabled = false)"
        if args.debug:
            print(cur.mogrify(sql))
        cur.execute(sql)
    print("\t... Success!")
    is_autovac_off = True
    cur.close()
    close_conn(vacuum_conn)


def vacuum_parent(conn, quoted_parent_table):
    cur = conn.cursor()
    sql = "VACUUM ANALYZE " + quoted_parent_table
    if args.debug:
        print(cur.mogrify(sql))
    if not args.quiet:
        print("Running vacuum analyze on parent table...")
    cur.execute(sql)
    cur.close()


if __name__ == "__main__":
    if args.version:
        print_version()

    if args.parent == None:
        print("-p/--parent option is required")
        sys.exit(2)

    if args.type == None:
        print("-t/--type option is required")
        sys.exit(2)

    is_autovac_off = False
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    conn = create_conn()
    partman_schema = get_partman_schema(conn)
    quoted_parent_table = get_quoted_parent_table(conn)

    try:
        if not args.autovacuum_on:
            turn_off_autovacuum(partman_schema, quoted_parent_table)

        total = partition_data(conn, partman_schema)

        if not args.quiet:
            print("Total rows moved: %d" % total)

        vacuum_parent(conn, quoted_parent_table)

    except:
        if not args.autovacuum_on and is_autovac_off == True:
            reset_autovacuum(partman_schema, quoted_parent_table)
        sys.exit(2)

    if not args.autovacuum_on:
        reset_autovacuum(partman_schema, quoted_parent_table)

    close_conn(conn)



#!/usr/bin/env python

import argparse, psycopg2, sys

partman_version = "2.3.2"

parser = argparse.ArgumentParser(description="Script for performing additional vacuum maintenance on the child tables of a partition set in order to avoid excess vacuuming and transaction id wraparound issues. Whether a table is vacuumed or not by this script depends on whether its age(relfrozenxid) is greater than vacuum_freeze_min_age. This ensures repeated runs of this script do not re-vacuum tables that have already had all of their page tuples frozen. This script does not require pg_partman to be installed and will work on any inheritance set given a parent table, as long as the options --all, --type or --interval are NOT set.")
parser.add_argument('-p', '--parent', help="Parent table of an already created partition/inheritance set. Either this option or --all is required.")
parser.add_argument('--all', action="store_true", help="Run against all tables managed by pg_partman. Either this option or -p is required. Overrides -p if both are set.")
parser.add_argument('-c', '--connection', default="host=", help="""Connection string for use by psycopg. Defaults to "host=" (local socket).""")
parser.add_argument('-z', '--freeze', action="store_true", help="Sets the FREEZE option to the VACUUM command.")
parser.add_argument('-f', '--full', action="store_true", help="Sets the FULL option to the VACUUM command. Note that --freeze is not necessary if you set this. Recommend reviewing --dryrun before running this since it will lock all tables it runs against, possibly including the parent.")
parser.add_argument('-a', '--vacuum_freeze_min_age', type=int, help="By default the script obtains this value from the system catalogs. By setting this, you can override the value obtained from the database. Note this does not change the value in the database, only the value this script uses.")
parser.add_argument('-t', '--type', choices=["time","id",], help="Tells the script whether the value given by --interval is either time or id based. Also, if --all is set, limits the partition sets that the script runs against to just those types. Valid values are: \"id\" and \"time\".")
parser.add_argument('-i', '--interval', help="In addition to checking transaction age, only run against tables that are older than the interval given. Takes either a postgresql time interval or integer value. For time-based partitioning, uses the time that script is run as baseline. For id-based partitioning, the integer value will set that any partitions with an id value less than the current maximum id value minus the given value will be vacuumed. For example, if the current max id is 100 and the given value is 30, any partitions with id values less than 70 will be vacuumed. The current maximum id value at the time the script is run is always used.")
parser.add_argument('--noparent', action="store_true", help="Normally the parent table is included in the list of tables to vacuum if its age(relfrozenxid) is higher than vacuum_freeze_min_age. Set this to force exclusion of the parent table, even if it meets that criteria.")
parser.add_argument('--dryrun', action="store_true", help="Show what the script will do without actually running it against the database. Highly recommend reviewing this before running for the first time.")
parser.add_argument('-q', '--quiet', action="store_true", help="Turn off all output.")
parser.add_argument('--version', action="store_true", help="Print out the minimum version of pg_partman this script is meant to work with. The version of pg_partman installed may be greater than this.")
parser.add_argument('--debug', action="store_true", help="Show additional debugging output")
args = parser.parse_args()


def close_conn(conn):
    conn.close()


def create_conn():
    conn = psycopg2.connect(args.connection)
    conn.autocommit = True
    return conn


def get_parent_table_list(conn):
    cur = conn.cursor()
    partman_schema = get_partman_schema(conn)
    sql = "SELECT parent_table FROM " + partman_schema + ".part_config"
    if args.type == "id":
        sql += " WHERE partition_type = 'id' "
    elif args.type == "time":
        sql += " WHERE partition_type = 'time' OR partition_type = 'time-custom' "
    sql += " ORDER BY parent_table"
    cur.execute(sql)
    parent_table_list = cur.fetchall()
    cur.close()
    return parent_table_list

def get_partition_list(conn, parent_table, vacuum_freeze_min_age):
    cur = conn.cursor()
    new_child_list = []
    # Put the parent table in the list so it gets frozen as well if necessary
    # Do not use show_partitions() so this can work on non-pg_partman tables
    sql = """
        WITH parent_info AS (
            SELECT c1.oid FROM pg_catalog.pg_class c1
            JOIN pg_catalog.pg_namespace n1 ON c1.relnamespace = n1.oid
            WHERE n1.nspname ||'.'|| c1.relname = %s
        )
        SELECT n.nspname::text, c.relname::text AS partition_name FROM
        pg_catalog.pg_inherits h
        JOIN pg_catalog.pg_class c ON c.oid = h.inhrelid
        JOIN pg_catalog.pg_namespace n ON c.relnamespace = n.oid
        JOIN parent_info pi ON h.inhparent = pi.oid
        ORDER BY 1,2"""
    if args.debug:
        print(cur.mogrify(sql, [parent_table]))
    cur.execute(sql, [parent_table])
    child_list = cur.fetchall()

    # only keep tables in list with age older than vacuum_freeze_min_age
    # this compare is much faster than --interval compare below, so reduce list size earlier
    for c in child_list:
        sql = """SELECT age(relfrozenxid) FROM pg_catalog.pg_class c JOIN pg_catalog.pg_namespace n ON c.relnamespace = n.oid
                WHERE n.nspname = %s
                AND c.relname = %s"""
        if args.debug:
            print(cur.mogrify(sql, [c[0], c[1]]))
        cur.execute(sql, [c[0], c[1]])
        table_age = cur.fetchone()[0]
        if args.debug:
            print("table age ("+c[1]+"): " + str(table_age))
        if table_age > vacuum_freeze_min_age:
            new_child_list.append(c)

    child_list = new_child_list

    # This may be used below with --interval as well, so just run it no matter what. Use fetchall() so tuple list is returned to match above list.
    sql = "SELECT schemaname, tablename FROM pg_catalog.pg_tables WHERE schemaname||'.'||tablename = %s"
    if args.debug:
        print(cur.mogrify(sql, [parent_table]))
    cur.execute(sql, [parent_table])
    parent_table_split = cur.fetchall()

    if args.interval != None:
        partman_schema = get_partman_schema(conn)
        new_child_list = []  #reset to reuse variable

        if args.type == "id":
            sql = "SELECT control FROM " + partman_schema + ".part_config WHERE parent_table = %s"
            cur.execute(sql, [parent_table])
            control_col = str(cur.fetchone()[0])
            sql = "SELECT max(\"" + control_col + "\") FROM \"" + str(parent_table_split[0][0]) + "\".\"" + str(parent_table_split[0][1]) + "\""
            cur.execute(sql)
            max_id = cur.fetchone()[0]
            upper_boundary = max_id - int(args.interval)
        elif args.type == "time":
            sql = "SELECT CURRENT_TIMESTAMP - %s::interval"  # Use CURRENT_TIMESTAMP since all variables in pg_partman are timestamptz
            cur.execute(sql, [args.interval])
            upper_boundary = cur.fetchone()[0]

        else:
            print("Invalid --type value passed into get_partition_list() function.")
            sys.exit(2)

        for c in child_list:

            if args.type == "id":
                sql = "SELECT child_start_id, child_end_id FROM " + partman_schema + ".show_partition_info(%s)"
                if args.debug:
                    print(cur.mogrify(sql, [c[0] + "." + c[1]]))
                cur.execute(sql, [c[0] + "." + c[1]])
                result = cur.fetchone()
                partition_min = result[0]
                partition_max = result[1]
                if args.debug:
                    print("upper_boundary: " + str(upper_boundary) + ", partition_min: " + str(partition_min) + " partition_max: " + str(partition_max))

                if partition_min < upper_boundary and partition_max < upper_boundary:
                    new_child_list.append(c)

            elif args.type == "time":
                sql = "SELECT child_start_time, child_end_time FROM " + partman_schema + ".show_partition_info(%s)"
                if args.debug:
                    print(cur.mogrify(sql, [c[0] + "." + c[1]]))
                cur.execute(sql, [c[0] + "." + c[1]])
                result = cur.fetchone()
                partition_min = result[0]
                partition_max = result[1]
                if args.debug:
                    print("upper_boundary: " + str(upper_boundary) + ", partition_min: " + str(partition_min) + " partition_max: " + str(partition_max))

                if partition_min < upper_boundary and partition_max < upper_boundary:
                    new_child_list.append(c)

        child_list = new_child_list

    if args.noparent == False:
        child_list += parent_table_split  # add parent table to existing list

    cur.close()
    return child_list


def get_partman_schema(conn):
    cur = conn.cursor()
    sql = "SELECT nspname FROM pg_catalog.pg_namespace n, pg_catalog.pg_extension e WHERE e.extname = 'pg_partman' AND e.extnamespace = n.oid"
    cur.execute(sql)
    partman_schema = "\"" + cur.fetchone()[0] + "\""
    cur.close()
    return partman_schema


def get_vacuum_freeze_min_age(conn):
    cur = conn.cursor()
    if args.vacuum_freeze_min_age == None:
        sql = "SELECT setting FROM pg_settings WHERE name = 'vacuum_freeze_min_age'"
        cur.execute(sql)
        vacuum_freeze_min_age = cur.fetchone()[0]
    else:
        vacuum_freeze_min_age = args.vacuum_freeze_min_age
    return int(vacuum_freeze_min_age)


def print_version():
    print(partman_version)
    sys.exit()


def vacuum_table(conn, schemaname, tablename):
    cur = conn.cursor()
    sql = "VACUUM "
    if args.full:
        sql += " FULL "
    if args.freeze:
        sql += " FREEZE "
    sql += " \"" + schemaname + "\".\"" +  tablename + "\" "
    if not args.quiet:
        print(sql)
    if not args.dryrun:
        cur.execute(sql)
    cur.close()


if __name__ == "__main__":
    if args.version:
        print_version()

    if args.parent == None and args.all == False:
        print("Either the -p/--parent option or --all must be set.")
        sys.exit(2)

    if args.parent != None and args.parent.find(".") < 0:
        print("Parent table must be schema qualified")
        sys.exit(2)

    if args.interval != None and args.type == None:
        print("--interval argment requires setting --type argument as well")
        sys.exit(2)

    main_conn = create_conn()
    vacuum_freeze_min_age = get_vacuum_freeze_min_age(main_conn)
    if args.debug:
        print("vacuum_freeze_min_age: " + str(vacuum_freeze_min_age))

    if args.all:
        parent_table_list = get_parent_table_list(main_conn)
    else:
        # put in nested list because above returns as one
        parent_table_list = [[args.parent]]

    for l in parent_table_list:
        partition_list = get_partition_list(main_conn, l[0], vacuum_freeze_min_age)
        for p in partition_list:
            vacuum_table(main_conn, p[0], p[1])

    close_conn(main_conn)



#!/usr/bin/env python

import argparse, psycopg2, signal, sys, time

partman_version = "4.0.0"

parser = argparse.ArgumentParser(description="This script calls either undo_partition(), undo_partition_time(), undo_partition_id(), or undo_partition_native() depending on the value given for --type. A commit is done at the end of each --interval and/or emptied partition. Returns the total number partitions and rows undone. Automatically stops when last child table is empty.", epilog="NOTE: To help avoid heavy load and contention during the undo, autovacuum is turned off for the entire partition set and all child tables when this script is run. When the undo is complete, autovacuum is set back to its default for the parent. Any child tables left behind will still have it turned off.")
parser.add_argument('-p','--parent', help="Parent table of the partition set. (Required)")
parser.add_argument('-t','--type', choices=["partman","native",], help="""Type of partitioning. Valid values are "partman" for old, trigger-based partition sets or "native" for natively partitioned sets. """)
parser.add_argument('-c','--connection', default="host=", help="""Connection string for use by psycopg. Defaults to "host=" (local socket). Note that two connections are required if allowing autovacuum to be turned off.""")
parser.add_argument('-i','--interval', help="Value that is passed on to the undo partitioning function as p_batch_interval. Use this to set an interval smaller than the partition interval to commit data in smaller batches. Defaults to the partition interval if not given. If -t value is not set, interval cannot be smaller than the partition interval and an entire partition is copied each batch.")
parser.add_argument('-b','--batch', type=int, default=0, help="How many times to loop through the value given for --interval. If --interval not set, will use default partition interval and undo at most -b partition(s).  Script commits at the end of each individual batch. (NOT passed as p_batch_count to undo function). If not set, all data will be moved to the parent table in a single run of the script.")
parser.add_argument('--native_target', help="""If unpartitioning a natively partitioned table, use this option to set the target table to send the data to. This option is required if -t is set to "native".""")
parser.add_argument('-d', '--droptable', action="store_true", help="Switch setting for whether to drop child tables when they are empty. Do not set to just uninherit.")
parser.add_argument('-w','--wait', type=float, default=0, help="Cause the script to pause for a given number of seconds between commits (batches).")
parser.add_argument('-l','--lockwait', default=0, type=float, help="Have a lock timeout of this many seconds on the data move. If a lock is not obtained, that batch will be tried again.")
parser.add_argument('--lockwait_tries', default=10, type=int, help="Number of times to allow a lockwait to time out before giving up on the partitioning.  Defaults to 10")
parser.add_argument('--autovacuum_on', action="store_true", help="Turning autovacuum off requires a brief lock to ALTER the table property. Set this option to leave autovacuum on and avoid the lock attempt.")
parser.add_argument('-q', '--quiet', action="store_true", help="Switch setting to stop all output during and after partitioning undo.")
parser.add_argument('--version', action="store_true", help="Print out the minimum version of pg_partman this script is meant to work with. The version of pg_partman installed may be greater than this.")
parser.add_argument('--debug', action="store_true", help="Show additional debugging output")
args = parser.parse_args() 

is_autovac_off = False

def close_conn(conn):
    conn.close()


def create_conn():
    conn = psycopg2.connect(args.connection)
    conn.autocommit = True
    return conn


def get_partman_schema(conn):
    cur = conn.cursor()
    query = "SELECT nspname FROM pg_catalog.pg_namespace n, pg_catalog.pg_extension e WHERE e.extname = 'pg_partman' AND e.extnamespace = n.oid"
    cur.execute(query)
    partman_schema = "\"" + cur.fetchone()[0] + "\""
    cur.close()
    return partman_schema


def get_quoted_parent_table(conn):
    cur = conn.cursor()
    query = "SELECT schemaname, tablename FROM pg_catalog.pg_tables WHERE schemaname||'.'||tablename = %s"
    cur.execute(query, [args.parent])
    result = cur.fetchone()
    if result == None:
        print("Given parent table ("+args.parent+") does not exist")
        sys.exit(2)
    quoted_parent_table = "\"" + result[0] + "\".\"" + result[1] + "\""
    cur.close()
    return quoted_parent_table


def print_version():
    print(partman_version)
    sys.exit()


def reset_autovacuum(partman_schema, quoted_parent_table):
    global is_autovac_off
    vacuum_conn = create_conn()
    cur = vacuum_conn.cursor()
    # Parameter doesn't exist on native parent tables
    if args.type == "partman":
        query = "ALTER TABLE " + quoted_parent_table + " RESET (autovacuum_enabled, toast.autovacuum_enabled)"
        if not args.quiet:
            print("Attempting to reset autovacuum for old parent table...")
        if args.debug:
            print(cur.mogrify(query))
        cur.execute(query)

    # Do not use show_partitions() so that this can work on non-pg_partman partition sets
    # query = "SELECT partition_schemaname, partition_tablename FROM " + partman_schema + ".show_partitions(%s)"
    query = """
        WITH parent_info AS (
            SELECT c1.oid FROM pg_catalog.pg_class c1
            JOIN pg_catalog.pg_namespace n1 ON c1.relnamespace = n1.oid
            WHERE n1.nspname ||'.'|| c1.relname = %s
        )
        SELECT n.nspname::text, c.relname::text AS partition_name FROM
        pg_catalog.pg_inherits h
        JOIN pg_catalog.pg_class c ON c.oid = h.inhrelid
        JOIN pg_catalog.pg_namespace n ON c.relnamespace = n.oid
        JOIN parent_info pi ON h.inhparent = pi.oid
        ORDER BY 1,2"""

    if args.debug:
        print(cur.mogrify(query, [args.parent]))
    cur.execute(query, [args.parent])
    result = cur.fetchall()
    for r in result:
        query = "ALTER TABLE \"" + r[0] + "\".\"" + r[1] + "\" RESET (autovacuum_enabled, toast.autovacuum_enabled)"
        if args.debug:
            print(cur.mogrify(query))
        cur.execute(query)
    print("\t... Success!")
    is_autovac_off = False
    cur.close()
    close_conn(vacuum_conn)


def signal_handler(signum, frame):
    if is_autovac_off == True:
        reset_autovacuum(conn, partman_schema, quoted_parent_table)
    sys.exit(2)

def turn_off_autovacuum(partman_schema, quoted_parent_table):
    global is_autovac_off
    vacuum_conn = create_conn()
    cur = vacuum_conn.cursor()
    # Parameter doesn't exist on native parent tables
    if args.type == "partman":
        query = "ALTER TABLE " + quoted_parent_table + " SET (autovacuum_enabled = false, toast.autovacuum_enabled = false)"
        if not args.quiet:
            print("Attempting to turn off autovacuum for partition set...")
        if args.debug:
            print(cur.mogrify(query))
        cur.execute(query)

    # Do not use show_partitions() so that this can work on non-pg_partman partition sets
    # query = "SELECT partition_schemaname, partition_tablename FROM " + partman_schema + ".show_partitions(%s)"
    query = """
        WITH parent_info AS (
            SELECT c1.oid FROM pg_catalog.pg_class c1
            JOIN pg_catalog.pg_namespace n1 ON c1.relnamespace = n1.oid
            WHERE n1.nspname ||'.'|| c1.relname = %s
        )
        SELECT n.nspname::text, c.relname::text AS partition_name FROM
        pg_catalog.pg_inherits h
        JOIN pg_catalog.pg_class c ON c.oid = h.inhrelid
        JOIN pg_catalog.pg_namespace n ON c.relnamespace = n.oid
        JOIN parent_info pi ON h.inhparent = pi.oid
        ORDER BY 1,2"""
    if args.debug:
        print(cur.mogrify(query, [args.parent]))
    cur.execute(query, [args.parent])
    result = cur.fetchall()
    for r in result:
        query = "ALTER TABLE \"" + r[0] + "\".\"" + r[1] + "\" SET (autovacuum_enabled = false, toast.autovacuum_enabled = false)"
        if args.debug:
            print(cur.mogrify(query))
        cur.execute(query)
    print("\t... Success!")
    is_autovac_off = True
    cur.close()
    close_conn(vacuum_conn)


def undo_partition_data(conn, partman_schema):
    batch_count = 0
    total_partitions = 0
    total_rows = 0
    lockwait_count = 0

    cur = conn.cursor()

    # Actual undo sql functions do not drop by default, so fix argument value to match that default
    if args.droptable:
        keep_table = False
    else:
        keep_table = True

    query = "SELECT partitions_undone, rows_undone FROM " + partman_schema + ".undo_partition"

    query += "(p_parent_table := %s, p_keep_table := %s, p_lock_wait := %s "
    if args.interval != None:
        query += ", p_batch_interval := %s "
    if args.type == "native":
        if args.native_target == None:
            print("""--native_target must be set if --type is set to "native" """)
            sys.exit(2)
        query += ", p_target_table := %s "
    query += ")"

    while True:
        #if args.interval != None:
        #    li = [args.parent, keep_table, args.interval, args.lockwait]
        #else:
        #    li = [args.parent, keep_table, args.lockwait]
        li = [args.parent, keep_table, args.lockwait]
        if args.interval != None:
            li.append(args.interval)
        if args.type == "native":
            li.append(args.native_target)
        if args.debug:
            print(cur.mogrify(query, li))
        cur.execute(query, li)

        result = cur.fetchone()
        if args.debug:
            print("Undo result: " + str(result))

        conn.commit()
        if not args.quiet:
            if result[0] > 0 or result[1] > 0:
                print("Partitions undone this batch: " + str(result[0]) + ". Rows undone this batch: " + str(result[1]))
            elif result[0] == -1:
                print("Unable to obtain lock, trying again (" + str(lockwait_count+1) + ")")
                print(conn.notices[-1])
        # if lock wait timeout, do not increment the counter
        if result[0] != -1:
            batch_count += 1
            total_partitions += result[0]
            total_rows += result[1]
            lockwait_count = 0
        else:
            lockwait_count += 1
            if lockwait_count > args.lockwait_tries:
                print("Quitting due to inability to get lock on table/rows for migration to parent")
                break
        print("Total rows moved: " + str(total_rows))
        # If given batch argument limit is reached
        if (args.batch > 0 and batch_count >= int(args.batch)):
            break
        # undo_partition functions will remove config entry once last child is dropped
        config_query = "SELECT parent_table FROM " + partman_schema + ".part_config WHERE parent_table = %s"
        if args.debug:
            print(cur.mogrify(config_query, [args.parent]))
        cur.execute(config_query, [args.parent])
        result = cur.fetchone()
        if result == None:
            break
        if args.wait > 0:
            time.sleep(args.wait)

    return total_partitions

def vacuum_parent(conn, quoted_parent_table):
    cur = conn.cursor()
    query = "VACUUM ANALYZE " + quoted_parent_table
    if args.debug:
        print(cur.mogrify(query))
    if not args.quiet:
        print("Running vacuum analyze on parent table...")
    cur.execute(query)
    cur.close()


if __name__ == "__main__":
    if args.version:
        print_version()

    if args.parent == None:
        print("-p/--parent option is required")
        sys.exit(2)

    if args.parent.find(".") < 0:
        print("ERROR: Parent table must be schema qualified")
        sys.exit(2)

    if args.type == "native" and args.native_target == None:
            print("""--native_target must be set if --type is set to "native" """)
            sys.exit(2)

    is_autovac_off = False
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    conn = create_conn()
    partman_schema = get_partman_schema(conn)
    quoted_parent_table = get_quoted_parent_table(conn)

    try:
        if not args.autovacuum_on:
            turn_off_autovacuum(partman_schema, quoted_parent_table)

        total = undo_partition_data(conn, partman_schema)

        if not args.quiet:
            print("Total partitions undone: %d" % total)

        vacuum_parent(conn, quoted_parent_table)

    except:
        if not args.autovacuum_on and is_autovac_off == True:
            reset_autovacuum(partman_schema, quoted_parent_table)
        raise

    if not args.autovacuum_on:
        reset_autovacuum(partman_schema, quoted_parent_table)

    close_conn(conn)



#!/usr/bin/env python

import argparse, psycopg2, sys, time
from multiprocessing import Process

partman_version = "2.2.0"

parser = argparse.ArgumentParser(description="Script for reapplying additional constraints managed by pg_partman on child tables. See docs for additional info on this special constraint management. Script runs in two distinct modes: 1) Drop all constraints  2) Apply all constraints. Typical usage would be to run the drop mode, edit the data, then run apply mode to re-create all constraints on a partition set.")
parser.add_argument('-p', '--parent', help="Parent table of an already created partition set. (Required)")
parser.add_argument('-c','--connection', default="host=", help="""Connection string for use by psycopg. Defaults to "host=" (local socket).""")
parser.add_argument('-d', '--drop_constraints', action="store_true", help="Drop all constraints managed by pg_partman. Drops constraints on all child tables including current & future.")
parser.add_argument('-a', '--add_constraints', action="store_true", help="Apply configured constraints to all child tables older than the optimize_constraint value.")
parser.add_argument('-j', '--jobs', type=int, default=0, help="Use the python multiprocessing library to recreate indexes in parallel. Value for -j is number of simultaneous jobs to run. Note that this is per table, not per index. Be very careful setting this option if load is a concern on your systems.")
parser.add_argument('-w', '--wait', type=float, default=0, help="Wait the given number of seconds after a table has had its constraints dropped or applied before moving on to the next. When used with -j, this will set the pause between the batches of parallel jobs instead.")
parser.add_argument('--dryrun', action="store_true", help="Show what the script will do without actually running it against the database. Highly recommend reviewing this before running.")
parser.add_argument('-q', '--quiet', action="store_true", help="Turn off all output.")
parser.add_argument('--version', action="store_true", help="Print out the minimum version of pg_partman this script is meant to work with. The version of pg_partman installed may be greater than this.")
args = parser.parse_args()


def apply_proc(child_table, partman_schema):
    conn = create_conn()
    conn.autocommit = True
    cur = conn.cursor()
    sql = "SELECT " + partman_schema + ".apply_constraints(%s, %s, %s, %s, %s)"
    debug = False;
    if not args.quiet:
        debug = True
        print(cur.mogrify(sql, [args.parent, child_table, False, None, debug]))
    if not args.dryrun:
        cur.execute(sql, [args.parent, child_table, False, None, debug])
    cur.close()
    close_conn(conn)


def create_conn():
    conn = psycopg2.connect(args.connection)
    return conn


def close_conn(conn):
    conn.close()


def drop_proc(child_table, partman_schema):
    conn = create_conn()
    conn.autocommit = True
    cur = conn.cursor()
    sql = "SELECT " + partman_schema + ".drop_constraints(%s, %s, %s)"
    debug = False;
    if not args.quiet:
        debug = True
        print(cur.mogrify(sql, [args.parent, child_table, debug]))
    if not args.dryrun:
        cur.execute(sql, [args.parent, child_table, debug])
    cur.close()
    close_conn(conn)


def get_children(conn, partman_schema):
    cur = conn.cursor()
    sql = "SELECT partition_schemaname||'.'||partition_tablename FROM " + partman_schema + ".show_partitions(%s, %s)"
    cur.execute(sql, [args.parent, 'ASC'])
    child_list = cur.fetchall()
    cur.close()
    return child_list


def get_partman_schema(conn):
    cur = conn.cursor()
    sql = "SELECT nspname FROM pg_catalog.pg_namespace n, pg_catalog.pg_extension e WHERE e.extname = 'pg_partman' AND e.extnamespace = n.oid"
    cur.execute(sql)
    partman_schema = "\"" + cur.fetchone()[0] + "\""
    cur.close()
    return partman_schema


def get_config_values(conn, partman_schema):
    # [0] = premake, [1] = optimize_constraint
    cur = conn.cursor()
    sql = "SELECT premake, optimize_constraint FROM " + partman_schema + ".part_config WHERE parent_table = %s"
    cur.execute(sql, [args.parent])
    config_values = cur.fetchone()
    cur.close()
    return config_values 


def get_quoted_parent_table(conn):
    cur = conn.cursor()
    sql = "SELECT schemaname, tablename FROM pg_catalog.pg_tables WHERE schemaname||'.'||tablename = %s"
    cur.execute(sql, [args.parent])
    result = cur.fetchone()
    if result == None:
        print("Given parent table ("+args.parent+") does not exist")
        sys.exit(2)
    quoted_parent_table = "\"" + result[0] + "\".\"" + result[1] + "\""
    cur.close()
    return quoted_parent_table


def print_version():
    print(partman_version)
    sys.exit()


if __name__ == "__main__":

    if args.version:
        print_version()

    if args.parent == None:
        print("-p/--parent option is required")
        sys.exit(2)

    if args.parent.find(".") < 0:
        print("Parent table must be schema qualified")
        sys.exit(2)

    if args.drop_constraints and args.add_constraints: 
        print("Can only set one or the other of --drop_constraints (-d) and --add_constraints (-a)")
        sys.exit(2)

    if (args.drop_constraints == False) and (args.add_constraints == False):
        print("Must set one of --drop_constraints (-d) or --add_constraints (-a)")
        sys.exit(2)

    main_conn = create_conn()
    partman_schema = get_partman_schema(main_conn)
    quoted_parent_table = get_quoted_parent_table(main_conn)
    child_list = get_children(main_conn, partman_schema)
    config_values = get_config_values(main_conn, partman_schema)
    premake = int(config_values[0])
    optimize_constraint = int(config_values[1])
    if args.add_constraints:
        # Remove tables from the list of child tables that shouldn't have constraints yet 
        for x in range(optimize_constraint + premake + 1):
            child_list.pop()

    if args.jobs == 0:
        for c in child_list:
            if args.drop_constraints:
               drop_proc(c[0], partman_schema)
            if args.add_constraints:
               apply_proc(c[0], partman_schema)
            if args.wait > 0:
                time.sleep(args.wait)
    else:
        child_list.reverse()
        while len(child_list) > 0:
            if not args.quiet:
                print("Jobs left in queue: " + str(len(child_list)))
            if len(child_list) < args.jobs:
                args.jobs = len(child_list)
            processlist = []
            for num in range(0, args.jobs):
                c = child_list.pop()
                if args.drop_constraints:
                    p = Process(target=drop_proc, args=(c[0], partman_schema))
                if args.add_constraints:
                    p = Process(target=apply_proc, args=(c[0], partman_schema))
                p.start()
                processlist.append(p)
            for j in processlist:
                j.join()
            if args.wait > 0:
                time.sleep(args.wait)

    sql = 'ANALYZE ' + quoted_parent_table
    main_cur = main_conn.cursor()
    if not args.quiet:
        print(main_cur.mogrify(sql))
    if not args.dryrun:
        main_cur.execute(sql)

    close_conn(main_conn)




#!/usr/bin/env python

import argparse, psycopg2, sys

partman_version = "3.1.0"

parser = argparse.ArgumentParser(description="This script will reapply the foreign keys on a parent table to all child tables in an inheritance set. Any existing foreign keys on child tables will be dropped in order to match the parent. A commit is done after each foreign key application to avoid excessive contention. Note that this script can work on any inheritance set, not just partition sets managed by pg_partman.")
parser.add_argument('-p','--parent', help="Parent table of an already created partition set. (Required)")
parser.add_argument('-c','--connection', default="host=", help="""Connection string for use by psycopg. Defaults to "host=" (local socket).""")
parser.add_argument('-q', '--quiet', action="store_true", help="Switch setting to stop all output during and after partitioning undo.")
parser.add_argument('--dryrun', action="store_true", help="Show what the script will do without actually running it against the database. Highly recommend reviewing this before running.")
parser.add_argument('--nonpartman', action="store_true", help="If the partition set you are running this on is not managed by pg_partman, set this flag. Otherwise internal pg_partman functions are used and this script may not work. When this is set the order that the tables are rekeyed is alphabetical instead of logical.")
parser.add_argument('--debug', action="store_true", help="Show additional debugging output")
parser.add_argument('--version', action="store_true", help="Print out the minimum version of pg_partman this script is meant to work with. The version of pg_partman installed may be greater than this.")
args = parser.parse_args()


def apply_foreign_keys(conn, child_tables, partman_schema):
    if not args.quiet:
        print("Applying foreign keys to child tables...")
    cur = conn.cursor()

    # Get template table if exists to use its indexes
    sql = "SELECT template_table FROM " + partman_schema + ".part_config WHERE parent_table = %s AND partition_type = 'native'"
    cur.execute(sql, [args.parent])
    template_table = cur.fetchone()

    for c in child_tables:
        sql = """SELECT pg_get_constraintdef(con.oid) AS constraint_def
                    FROM pg_catalog.pg_constraint con
                    JOIN pg_catalog.pg_class c ON con.conrelid = c.oid
                    JOIN pg_catalog.pg_namespace n ON c.relnamespace = n.oid
                    WHERE n.nspname ||'.'|| c.relname = %s
                    AND contype = 'f'""";

        if template_table == None:
            if args.debug:
                print(cur.mogrify(sql, [args.parent]))
            cur.execute(sql, [args.parent])
        else:
            if args.debug:
                print(cur.mogrify(sql, [template_table]))
            cur.execute(sql, [template_table])

        parent_fkeys = cur.fetchall()
        for pfk in parent_fkeys:
            alter_sql = "ALTER TABLE \"" + c[0] + "\".\"" + c[1] + "\" ADD " + pfk[0]

            if not args.quiet:
                print(alter_sql)
            if not args.dryrun:
                cur.execute(alter_sql)


def create_conn():
    conn = psycopg2.connect(args.connection)
    conn.autocommit = True
    return conn


def close_conn(conn):
    conn.close()


def drop_foreign_keys(conn, child_tables):
    if not args.quiet:
        print("Dropping current foreign keys on child tables...")
    cur = conn.cursor()
    for c in child_tables:
        sql = """SELECT constraint_name
            FROM information_schema.table_constraints
            WHERE table_schema = %s AND table_name = %s AND constraint_type = 'FOREIGN KEY'"""
        if args.debug:
            print(cur.mogrify(sql, [ c[0], c[1] ]))
        cur.execute(sql, [ c[0], c[1] ])
        child_fkeys = cur.fetchall()
        for cfk in child_fkeys:
            alter_sql = "ALTER TABLE \"" + c[0] + "\".\"" + c[1] + "\" DROP CONSTRAINT \"" + cfk[0] + "\""
            if not args.quiet:
                print(alter_sql)
            if not args.dryrun:
                cur.execute(alter_sql)


def get_child_tables(conn, partman_schema):
    if not args.quiet:
        print("Getting list of child tables...")
    cur = conn.cursor()
    if args.nonpartman == False:
        sql = "SELECT partition_schemaname, partition_tablename FROM " + partman_schema + ".show_partitions(%s)"
    else:
        sql = """
            WITH parent_info AS (
                SELECT c1.oid FROM pg_catalog.pg_class c1
                JOIN pg_catalog.pg_namespace n1 ON c1.relnamespace = n1.oid
                WHERE n1.nspname ||'.'|| c1.relname = %s
            )
            SELECT n.nspname::text, c.relname::text AS partition_name FROM
            pg_catalog.pg_inherits h
            JOIN pg_catalog.pg_class c ON c.oid = h.inhrelid
            JOIN pg_catalog.pg_namespace n ON c.relnamespace = n.oid
            JOIN parent_info pi ON h.inhparent = pi.oid
            ORDER BY 1,2"""
    if args.debug:
        print(cur.mogrify(sql, [args.parent]))
    cur.execute(sql, [args.parent])
    result = cur.fetchall()
    return result


def get_partman_schema(conn):
    cur = conn.cursor()
    sql = "SELECT nspname FROM pg_catalog.pg_namespace n, pg_catalog.pg_extension e WHERE e.extname = 'pg_partman' AND e.extnamespace = n.oid"
    cur.execute(sql)
    partman_schema = "\"" + cur.fetchone()[0] + "\""
    cur.close()
    return partman_schema


def print_version():
    print(partman_version)
    sys.exit()


if __name__ == "__main__":

    if args.version:
        print_version()

    if args.parent == None:
        print("-p/--parent option is required")
        sys.exit(2)

    conn = create_conn()

    partman_schema = get_partman_schema(conn)
    child_tables = get_child_tables(conn, partman_schema)

    drop_foreign_keys(conn, child_tables)
    apply_foreign_keys(conn, child_tables, partman_schema)

    if not args.quiet:
        print("Done!")
    close_conn(conn)



#!/usr/bin/env python

import argparse, psycopg2, re, sys, time
from multiprocessing import Process

partman_version = "4.1.0"

parser = argparse.ArgumentParser(description="Script for reapplying indexes on child tables in a partition set to match the parent table. Any indexes that currently exist on the children and match the definition on the parent will be left as is. There is an option to recreate matching as well indexes if desired, as well as the primary key. Indexes that do not exist on the parent will be dropped. Commits are done after each index is dropped/created to help prevent long running transactions & locks.", epilog="NOTE: New index names are made based off the child table name & columns used, so their naming may differ from the name given on the parent. This is done to allow the tool to account for long or duplicate index names. If an index name would be duplicated, an incremental counter is added on to the end of the index name to allow it to be created. Use the --dryrun option first to see what it will do and which names may cause dupes to be handled like this.")
parser.add_argument('-p', '--parent', help="Parent table of an already created partition set. (Required)")
parser.add_argument('-c','--connection', default="host=", help="""Connection string for use by psycopg. Defaults to "host=" (local socket).""")
parser.add_argument('--concurrent', action="store_true", help="Create indexes with the CONCURRENTLY option. Note this does not work on primary keys when --primary is given.")
parser.add_argument('--drop_concurrent', action="store_true", help="If an index is dropped (because it doesn't exist on the parent or because you set them to be recreated), do it concurrently. Note this does not work on primary keys when --primary is given.")
parser.add_argument('-R', '--recreate_all', action="store_true", help="By default, if an index exists on a child and matches the parent, it will not be touched. Setting this option will force all child indexes to be dropped & recreated. Will obey the --concurrent & --drop_concurrent options if given. Will not recreate primary keys unless --primary option is also given.")
parser.add_argument('--primary', action="store_true", help="By default the primary key is not recreated. Set this option if that is needed. Note this will cause an exclusive lock on the child table.")
parser.add_argument('-j', '--jobs', type=int, default=0, help="Use the python multiprocessing library to recreate indexes in parallel. Value for -j is number of simultaneous jobs to run. Note that this is per table, not per index. Be very careful setting this option if load is a concern on your systems.")
parser.add_argument('-w', '--wait', type=float, default=0, help="Wait the given number of seconds after indexes have finished being created on a table before moving on to the next. When used with -j, this will set the pause between the batches of parallel jobs instead.")
parser.add_argument('--dryrun', action="store_true", help="Show what the script will do without actually running it against the database. Highly recommend reviewing this before running. Note that if multiple indexes would get the same default name, the duplicated name will show in the dryrun (because the index doesn't exist in the catalog to check for it). When the real thing is run, the duplicated names will be handled as stated in NOTE at the end of --help.")
parser.add_argument('-q', '--quiet', action="store_true", help="Turn off all output.")
parser.add_argument('--nonpartman', action="store_true", help="If the partition set you are running this on is not managed by pg_partman, set this flag otherwise this script may not work. Note that the pg_partman extension is still required to be installed for this to work since it uses certain internal functions. When this is set the order that the tables are reindexed is alphabetical instead of logical.")
parser.add_argument('--version', action="store_true", help="Print out the minimum version of pg_partman this script is meant to work with. The version of pg_partman installed may be greater than this.")
args = parser.parse_args()

def check_compatibility(conn, partman_schema):
    cur = conn.cursor()
    
    sql = """SELECT current_setting('server_version_num')::int"""
    cur.execute(sql)
    pg_version = int(cur.fetchone()[0])

    if pg_version < 90400:
        print("ERROR: This script requires PostgreSQL minimum version of 9.4.0")
        sys.exit(2)

    sql = "SELECT partition_type FROM " + partman_schema + ".part_config WHERE parent_table = %s"
    cur.execute(sql, [args.parent])
    partition_type = cur.fetchone()[0]

    if pg_version >= 110000 and partition_type == "native":
        print("This script cannot currently work with native partition sets in PG11+. Please use native index inheritance methods if possible.")
        cur.close()
        close_conn(conn)
        sys.exit(2)

    cur.close()


def create_conn():
    conn = psycopg2.connect(args.connection)
    return conn


def create_index(conn, partman_schema, child_schemaname, child_tablename, child_index_list, parent, parent_index_list):
    cur = conn.cursor()
    sql = """SELECT schemaname, tablename FROM pg_catalog.pg_tables WHERE schemaname||'.'||tablename = %s"""
    cur.execute(sql, [parent])
    result = cur.fetchone()
    parent_schemaname = result[0]
    parent_tablename = result[1]
    cur.close()
    parent_match_regex = re.compile(r" ON \"?%s\"?\.\"?%s\"? | ON \"?%s\"?" % (parent_schemaname, parent_tablename, parent_tablename))
    index_match_regex = re.compile(r'(?P<index_def>USING .*)')
    for i in parent_index_list:
        # if there is already a child index that matches the parent index don't try to create it unless --recreate_all set
        if args.recreate_all != True:
            child_found = False
            statement = None
            parinddef = index_match_regex.search(i[0])
            for c in child_index_list:
                chinddef = index_match_regex.search(c[0])
                if chinddef.group('index_def') == parinddef.group('index_def'):
                    child_found = True
                    break
            if child_found:
                continue

        if i[1] == True and args.primary:
            index_name = child_tablename + "_" + "_".join(i[2].split(","))
            sql = "SELECT " + partman_schema + ".check_name_length('" + index_name + "', p_suffix := '_pk')"
            cur = conn.cursor()
            cur.execute(sql)
            index_name = cur.fetchone()[0]
            cur.close()
            quoted_column_names = "\"" + "\",\"".join(i[2].split(",")) + "\""
            statement = "ALTER TABLE \"" + child_schemaname + "\".\"" + child_tablename + "\" ADD CONSTRAINT \"" + index_name + "\" PRIMARY KEY (" + quoted_column_names + ")"
        elif i[1] == False:
            index_name = child_tablename
            if i[2] != None:
                index_name += "_"
                index_name += "_".join(i[2].split(","))
            sql = "SELECT " + partman_schema + ".check_name_length('" + index_name + "', p_suffix := '_idx')"
            cur = conn.cursor()
            cur.execute(sql)
            index_name = cur.fetchone()[0]
            name_counter = 1
            while True:
                sql = "SELECT count(*) FROM pg_class c JOIN pg_namespace n ON c.relnamespace = n.oid WHERE n.nspname = %s AND c.relname = %s"
                cur = conn.cursor()
                cur.execute(sql, [child_schemaname, index_name])
                index_exists = cur.fetchone()[0]
                if index_exists != None and index_exists > 0:
                    index_name = child_tablename
                    if i[2] != None:
                        index_name += "_"
                        index_name += "_".join(i[2].split(","))
                    suffix = "_idx" + str(name_counter)
                    sql = "SELECT " + partman_schema + ".check_name_length('" + index_name + "', p_suffix := '" + suffix + "')"
                    cur = conn.cursor()
                    cur.execute(sql)
                    index_name = cur.fetchone()[0]
                    name_counter += 1
                else:
                    break
            cur.close()
            statement = i[0]
            statement = statement.replace(i[3], index_name)  # replace parent index name with child index name
            if args.concurrent:
                statement = statement.replace("CREATE INDEX ", "CREATE INDEX CONCURRENTLY ")
            statement = parent_match_regex.sub(" ON \"" + child_schemaname + "\".\"" + child_tablename + "\" ", statement)
        cur = conn.cursor()
        if statement != None:
            if not args.quiet:
                print(cur.mogrify(statement))
            if not args.dryrun:
                cur.execute(statement)
        cur.close()


def close_conn(conn):
    conn.close()


def drop_index(conn, child_schemaname, child_tablename, child_index_list, parent_index_list):
    cur = conn.cursor()
    for d in child_index_list:
        if d[1] == True and args.primary:
            statement = "ALTER TABLE \"" + child_schemaname + "\".\"" + child_tablename + "\" DROP CONSTRAINT \"" + d[4] + "\""
            if not args.quiet:
                print(cur.mogrify(statement))
            if not args.dryrun:
                cur.execute(statement)
        elif d[1] == False:
            if args.drop_concurrent:
                statement = "DROP INDEX CONCURRENTLY \"" + d[2] + "\".\"" + d[3] + "\""
            else:
                statement = "DROP INDEX \"" + d[2] + "\".\"" + d[3] + "\""

            if args.recreate_all != True:
                pat = re.compile(r'(?P<index_def>USING .*)')
                # if there is a parent index that matches the child index
                # don't try to drop it - we'd just have to recreate it
                chinddef = pat.search(d[0])
                parent_found = False
                for p in parent_index_list:
                    parent_found = False
                    parinddef = pat.search(p[0])
                    if chinddef.group('index_def') == parinddef.group('index_def'):
                        parent_found = True
                        break
                if not parent_found:
                    if not args.quiet:
                        print(cur.mogrify(statement))
                    if not args.dryrun:
                        cur.execute(statement)
            else:
                if not args.quiet:
                    print(cur.mogrify(statement))
                if not args.dryrun:
                    cur.execute(statement)


def get_children(conn, partman_schema):
    cur = conn.cursor()
    if args.nonpartman == False:
        sql = "SELECT partition_schemaname, partition_tablename FROM " + partman_schema + ".show_partitions(%s)"
    else:
        sql = """
            WITH parent_info AS (
                SELECT c1.oid FROM pg_catalog.pg_class c1
                JOIN pg_catalog.pg_namespace n1 ON c1.relnamespace = n1.oid
                WHERE n1.nspname ||'.'|| c1.relname = %s
            )
            SELECT n.nspname::text, c.relname::text AS partition_name FROM
            pg_catalog.pg_inherits h
            JOIN pg_catalog.pg_class c ON c.oid = h.inhrelid
            JOIN pg_catalog.pg_namespace n ON c.relnamespace = n.oid
            JOIN parent_info pi ON h.inhparent = pi.oid
            ORDER BY 1,2"""
    cur.execute(sql, [args.parent])
    child_list = cur.fetchall()
    cur.close()
    return child_list


def get_child_index_list(conn, child_schemaname, child_tablename):
    cur = conn.cursor()
    sql = """
            WITH child_info AS (
                SELECT c1.oid FROM pg_catalog.pg_class c1
                JOIN pg_catalog.pg_namespace n1 ON c1.relnamespace = n1.oid
                WHERE n1.nspname = %s
                AND c1.relname = %s
            )
            SELECT pg_get_indexdef(indexrelid) AS statement
            , i.indisprimary
            , n.nspname AS index_schemaname
            , c.relname AS index_name
            , t.conname
            FROM pg_catalog.pg_index i
            JOIN pg_catalog.pg_class c ON i.indexrelid = c.oid
            JOIN pg_catalog.pg_namespace n ON c.relnamespace = n.oid
            LEFT JOIN pg_catalog.pg_constraint t ON c.oid = t.conindid
            JOIN child_info ci ON i.indrelid = ci.oid """
    cur.execute(sql, [child_schemaname, child_tablename])
    child_index_list = cur.fetchall()
    cur.close()
    return child_index_list


def get_parent_index_list(conn, parent):
    cur = conn.cursor()

    sql = """
            WITH parent_info AS (
                SELECT c1.oid FROM pg_catalog.pg_class c1
                JOIN pg_catalog.pg_namespace n1 ON c1.relnamespace = n1.oid
                WHERE n1.nspname||'.'||c1.relname = %s
            )
            SELECT
            pg_get_indexdef(indexrelid) AS statement
            , i.indisprimary
            , ( SELECT array_to_string(array_agg( a.attname ORDER by x.r ), ',')
                FROM pg_catalog.pg_attribute a
                JOIN ( SELECT k, row_number() over () as r
                        FROM unnest(i.indkey) k ) as x
                ON a.attnum = x.k AND a.attrelid = i.indrelid
            ) AS indkey_names
            , c.relname AS index_name
            FROM pg_catalog.pg_index i
            JOIN pg_catalog.pg_class c ON i.indexrelid = c.oid
            JOIN parent_info pi ON i.indrelid = pi.oid
            WHERE i.indisvalid
            ORDER BY 1"""
    cur.execute(sql, [parent])
    parent_index_list = cur.fetchall()
    return parent_index_list


def get_parent(conn, partman_schema):
    cur = conn.cursor()

    sql = "SELECT template_table FROM " + partman_schema + ".part_config WHERE parent_table = %s AND partition_type = 'native'"
    cur.execute(sql, [args.parent])
    template_table = cur.fetchone()

    if template_table is None:
        return args.parent
    else:
        return template_table[0]


def get_partman_schema(conn):
    cur = conn.cursor()
    sql = "SELECT nspname FROM pg_catalog.pg_namespace n, pg_catalog.pg_extension e WHERE e.extname = 'pg_partman' AND e.extnamespace = n.oid"
    cur.execute(sql)
    partman_schema = "\"" + cur.fetchone()[0] + "\""
    cur.close()
    return partman_schema


def print_version():
    print(partman_version)
    sys.exit()


def reindex_proc(child_schemaname, child_tablename, parent, parent_index_list, partman_schema):
    conn = create_conn()
    conn.autocommit = True # must be turned on to support CONCURRENTLY
    cur = conn.cursor()
    child_index_list = get_child_index_list(conn, child_schemaname, child_tablename)
    drop_index(conn, child_schemaname, child_tablename, child_index_list, parent_index_list)
    create_index(conn, partman_schema, child_schemaname, child_tablename, child_index_list, parent, parent_index_list)

    sql = "ANALYZE \"" + child_schemaname + "\".\"" + child_tablename + "\""
    if not args.quiet:
        print(cur.mogrify(sql))
    if not args.dryrun:
        cur.execute(sql)
    cur.close()
    close_conn(conn)


if __name__ == "__main__":

    if args.version:
        print_version()

    if args.parent == None:
        print("-p/--parent option is required")
        sys.exit(2)

    if args.parent.find(".") < 0:
        print("ERROR: Parent table must be schema qualified")
        sys.exit(2)

    conn = create_conn()

    partman_schema = get_partman_schema(conn)
    check_compatibility(conn,partman_schema)
    parent = get_parent(conn, partman_schema)
    parent_index_list = get_parent_index_list(conn, parent)
    child_list = get_children(conn, partman_schema)
    close_conn(conn)

    if args.jobs == 0:
         for c in child_list:
            reindex_proc(c[0], c[1], parent, parent_index_list, partman_schema)
            if args.wait > 0:
                time.sleep(args.wait)
    else:
        child_list.reverse()
        while len(child_list) > 0:
            if not args.quiet:
                print("Jobs left in queue: " + str(len(child_list)))
            if len(child_list) < args.jobs:
                args.jobs = len(child_list)
            processlist = []
            for num in range(0, args.jobs):
                c = child_list.pop()
                p = Process(target=reindex_proc, args=(c[0], c[1], parent, parent_index_list, partman_schema))
                p.start()
                processlist.append(p)
            for j in processlist:
                j.join()
            if args.wait > 0:
                time.sleep(args.wait)



#!/usr/bin/env python

import argparse, collections, psycopg2, sys, tempfile

partman_version = "2.0.0"

parser = argparse.ArgumentParser(description="This script is used to check that all rows in a partition set are unique for the given columns. Since unique constraints are not applied across partition sets, this cannot be enforced within the database. This script can be used as a monitor to ensure uniquness. If any unique violations are found, the values, along with a count of each, are output.")
parser.add_argument('-p', '--parent',  help="Parent table of the partition set to be checked")
parser.add_argument('-l', '--column_list', help="Comma separated list of columns that make up the unique constraint to be checked")
parser.add_argument('-c','--connection', default="host=", help="""Connection string for use by psycopg. Defaults to "host=" (local socket).""")
parser.add_argument('-t', '--temp', help="Path to a writable folder that can be used for temp working files. Defaults system temp folder.")
parser.add_argument('--psql', help="Full path to psql binary if not in current PATH")
parser.add_argument('--simple', action="store_true", help="Output a single integer value with the total duplicate count. Use this for monitoring software that requires a simple value to be checked for.")
parser.add_argument('--index_scan', action="store_true", help="By default index scans are disabled to force the script to check the actual table data with sequential scans. Set this option if you want the script to allow index scans to be used (does not guarentee that they will be used).")
parser.add_argument('-q', '--quiet', action="store_true", help="Suppress all output unless there is a constraint violation found.")
parser.add_argument('--version', action="store_true", help="Print out the minimum version of pg_partman this script is meant to work with. The version of pg_partman installed may be greater than this.")
args = parser.parse_args()

if args.version:
    print(partman_version)
    sys.exit()

if args.parent == None:
    print("-p/--parent option is required")
    sys.exit(2)

if args.column_list == None:
    print("-l/--column_list option is required")
    sys.exit(2)

if args.temp == None:
    tmp_copy_file = tempfile.NamedTemporaryFile(prefix="partman_constraint")
else:
    tmp_copy_file = tempfile.NamedTemporaryFile(prefix="partman_constraint", dir=args.temp)

fh = open(tmp_copy_file.name, 'w')
conn = psycopg2.connect(args.connection)
conn.set_session(isolation_level="REPEATABLE READ", readonly=True)
cur = conn.cursor()

# Get separate object names so they can be properly quoted
sql = "SELECT schemaname, tablename FROM pg_catalog.pg_tables WHERE schemaname||'.'||tablename = %s"
cur.execute(sql, [args.parent])
result = cur.fetchone()
if result == None:
    print("Given parent table ("+args.parent+") does not exist")
    sys.exit(2)
quoted_parent_table = "\"" + result[0] + "\".\"" + result[1] + "\""
#print(quoted_parent_table)

# Recreated column list with double quotes around each
cols_array = args.column_list.split(",")
quoted_col_list = "\"" + "\",\"".join(cols_array) + "\""
#print(quoted_col_list)

if args.index_scan == False:
    sql = """set enable_bitmapscan = false;
    set enable_indexonlyscan = false;
    set enable_indexscan = false;
    set enable_seqscan = true;"""
else:
    sql = """set enable_bitmapscan = true;
    set enable_indexonlyscan = true;
    set enable_indexscan = true;
    set enable_seqscan = false;"""
cur.execute(sql)
cur.close()
cur = conn.cursor()
if not args.quiet:
    print("Dumping out column data to temp file...")
copy_statement = "COPY (SELECT " + quoted_col_list + " FROM " + quoted_parent_table + ") TO STDOUT WITH DELIMITER ','"
#print(copy_statement)
cur.copy_expert(copy_statement, fh)
conn.rollback()
conn.close()
fh.close()

total_count = 0
if not args.quiet:
    print("Checking for dupes...")
with open(tmp_copy_file.name) as infile:
    counts = collections.Counter(l.strip() for l in infile)
for line, count in counts.most_common():
    if count > 1:
        if not args.simple:
            print(str(line) + ": " + str(count))
        total_count += count

if args.simple:
    if total_count > 0:
        print(total_count)
    elif not args.quiet:
        print(total_count)
else:
    if total_count == 0 and not args.quiet:
        print("No constraint violations found")



#!/usr/bin/env python

import argparse, hashlib, os, os.path, psycopg2, subprocess, sys

partman_version = "2.0.0"

parser = argparse.ArgumentParser(description="This script will dump out and then drop all tables contained in the designated schema using pg_dump.  Each table will be in its own separate file along with a SHA-512 hash of the dump file. Tables are not dropped from the database if pg_dump does not return successfully. All dump_* option defaults are the same as they would be for pg_dump if they are not given.", epilog="NOTE: The connection options for psyocpg and pg_dump were separated out due to distinct differences in their requirements depending on your database connection configuration.")
parser.add_argument('-n','--schema', help="(Required) The schema that contains the tables that will be dumped.")
parser.add_argument('-c','--connection', default="host=", help="""Connection string for use by psycopg. Role used must be able to select pg_catalog.pg_tables in the relevant database and drop all tables in the given schema.  Defaults to "host=" (local socket). Note this is distinct from the parameters sent to pg_dump.""")
parser.add_argument('-o','--output', default=os.getcwd(), help="Path to dump file output location. Default is where the script is run from.")
parser.add_argument('-d','--dump_database', help="Used for pg_dump, same as its --dbname (-d) option or final database name parameter.")
parser.add_argument('--dump_host', help="Used for pg_dump, same as its --host (-h) option.")
parser.add_argument('--dump_username', help="Used for pg_dump, same as its --username (-U) option.")
parser.add_argument('--dump_port', help="Used for pg_dump, same as its --port (-p) option.")
parser.add_argument('--pg_dump_path', help="Path to pg_dump binary location. Must set if not in current PATH.")
parser.add_argument('--Fp', action="store_true", help="Dump using pg_dump plain text format. Default is binary custom (-Fc).")
parser.add_argument('--nohashfile', action="store_true", help="Do NOT create a separate file with the SHA-512 hash of the dump. If dump files are very large, hash generation can possibly take a long time.")
parser.add_argument('--nodrop', action="store_true", help="Do NOT drop the tables from the given schema after dumping/hashing.")
parser.add_argument('-v','--verbose', action="store_true", help="Provide more verbose output.")
parser.add_argument('--version', action="store_true", help="Print out the minimum version of pg_partman this script is meant to work with. The version of pg_partman installed may be greater than this.")
args = parser.parse_args()


def create_hash(table_name):
    output_file = os.path.join(args.output, args.schema + "." + table_name + ".pgdump")
    try:
        with open(output_file, "rb") as fh:
            shash = hashlib.sha512()
            while True:
                data = fh.read(8192)
                if not data:
                    break
                shash.update(data)
    except IOError as e:
        print("Cannot access dump file for hash creation: " + e.strerror)
        sys.exit(2)

    hash_file = os.path.join(args.output, args.schema + "." + table_name + ".hash")
    if args.verbose:
        print("hash_file: " + hash_file)
    try:
        with open(hash_file, "w") as fh:
            fh.write(shash.hexdigest() + "  " + os.path.basename(output_file))
    except IOError as e:
        print("Unable to write to hash file: " + e.strerror)
        sys.exit(2)


def drop_table(table_name):
    conn = psycopg2.connect(args.connection)
    cur = conn.cursor()
    sql = "DROP TABLE IF EXISTS \"" + args.schema + "\".\"" + table_name + "\"";
    print(sql)
    cur.execute(sql)
    conn.commit()
    cur.close()
    conn.close()


def get_tables():
    conn = psycopg2.connect(args.connection)
    cur = conn.cursor()
    sql = "SELECT tablename FROM pg_catalog.pg_tables WHERE schemaname = %s";
    cur.execute(sql, [args.schema])
    result = cur.fetchall()
    cur.close()
    conn.close()
    return result


def perform_dump(result):
    table_name = result.pop()[0]
    processcmd = []
    if args.pg_dump_path != None:
        processcmd.append(args.pg_dump_path)
    else:
        processcmd.append("pg_dump")
    if args.dump_host != None:
        processcmd.append("--host=" + args.dump_host)
    if args.dump_port != None:
        processcmd.append("--port=" + args.dump_port)
    if args.dump_username != None:
        processcmd.append("--username=" + args.dump_username)
    if args.Fp:
        processcmd.append("--format=plain")
    else:
        processcmd.append("--format=custom")
    processcmd.append("--table=\"" + args.schema + "\".\"" + table_name + "\"")
    output_file = os.path.join(args.output, args.schema + "." + table_name + ".pgdump")
    processcmd.append("--file=" + output_file)
    if args.dump_database != None:
        processcmd.append(args.dump_database)

    if args.verbose:
        print(processcmd)
    try:
        subprocess.check_call(processcmd)
    except subprocess.CalledProcessError as e:
        print("Error in pg_dump command: " + str(e.cmd))
        sys.exit(2)

    return table_name 


def print_version():
    print(partman_version)
    sys.exit()


if __name__ == "__main__":

    if args.version:
        print_version()

    if args.schema == None:
        print("-n/--schema option is required")
        sys.exit(2)

    if not os.path.exists(args.output):
        print("Path given by --output (-o) does not exist: " + str(args.output))
        sys.exit(2)

    result = get_tables()
    while len(result) > 0:
        table_name = perform_dump(result)
        if not args.nohashfile:
            create_hash(table_name)
        if not args.nodrop:
            drop_table(table_name)
